#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <inttypes.h>
#include "../sf_model.h"
#include "../check_syscalls.h"
#include "../io_helpers.h"

int64_t num_models=0;
struct sf_model_allz *models = NULL;
struct sf_model_allz min={{0}}, max={{0}};

void calc_hist(int64_t j, FILE *output);
void calc_dist(int64_t j, int64_t k, FILE *output);

int main(int argc, char **argv) {
  int64_t i,j,k;
  char buffer[2048];
  if (argc < 2) {
    fprintf(stderr, "Usage: %s mcmc_output.dat\n", argv[0]);
    exit(EXIT_FAILURE);
  }
  
  FILE *in = check_fopen(argv[1], "r");
  struct sf_model_allz m = {{0}};
  check_mkdir("pdfs", 0755);
  while (fgets(buffer, 2048, in)) {
    if (read_params(buffer, m.params, NUM_PARAMS+1)<(NUM_PARAMS+1)) continue;
#define translate2(a,b) (a)-=(b)
#define translate3(a,b,c) (a)-=((b)+(c))
    translate3(EFF_0_A(m), EFF_0_A2(m), EFF_0_A3(m));
    translate3(V_1_A(m), V_1_A2(m), V_1_A3(m));
    translate3(ALPHA_A(m), ALPHA_A2(m), ALPHA_A3(m));
    translate2(BETA_A(m), BETA_A2(m));
    translate2(GAMMA_A(m), GAMMA_A2(m));
    translate2(Q_LVMP_A(m), Q_LVMP_Z(m));
    translate2(Q_SIG_LVMP_A(m), Q_SIG_LVMP_Z(m));
    DUST_WIDTH(m) = 0.23/DUST_WIDTH(m);
    DUST_M(m) = -20-2.5*(DUST_M(m)-9.2);
    DUST_M_Z(m) = -2.5*DUST_M_Z(m);
    check_realloc_every(models, sizeof(struct sf_model_allz), num_models, 1000);
    models[num_models] = m;
    if (num_models==0) { min=m; max=m; }
    for (i=0; i<NUM_PARAMS+1; i++) {
      if (m.params[i] < min.params[i]) min.params[i] = m.params[i];
      if (m.params[i] > max.params[i]) max.params[i] = m.params[i];
    }
    num_models++;
  }
  fclose(in);

  FILE *output = check_fopen("mcmc_pdf_stats.dat", "w");
  for (j=0; j<NUM_PARAMS+1; j++) {
    for (k=0; k<NUM_PARAMS+1; k++) {
      if (j==k) calc_hist(j, output);
      else calc_dist(j,k,output);
    }
  }
  fclose(output);
  return 0;
}


#define NUM_HIST_BINS 25
void calc_hist(int64_t j, FILE *output) {
  double bins[NUM_HIST_BINS] = {0};
  char buffer[1024];
  int64_t i;
  if (max.params[j]==min.params[j]) return;
  double upbin = (double)NUM_HIST_BINS/(max.params[j]-min.params[j]);
  for (i=0; i<num_models; i++) {
    int64_t bin = (models[i].params[j]-min.params[j])*upbin;
    if (bin>=NUM_HIST_BINS) bin=NUM_HIST_BINS-1;
    bins[bin]++;
  }
  snprintf(buffer, 1024, "pdfs/hist_%"PRId64".dat", j);
  FILE *data = check_fopen(buffer, "w");
  for (i=0; i<NUM_HIST_BINS; i++) {
    bins[i] *= upbin/(double)num_models;
    fprintf(data, "%lf %lf\n", min.params[j] + (i+0.5)*(max.params[j]-min.params[j])/(double)NUM_HIST_BINS, bins[i]);
  }
  fclose(data);

  fprintf(output, "%s %lf %lf %lf %lf\n", buffer, min.params[j], max.params[j], (min.params[j]+0.5/upbin), (max.params[j]-0.5/upbin));
}

#define NUM_X_BINS 40
#define NUM_Y_BINS 30
void calc_dist(int64_t j, int64_t k, FILE *output) {
  double bins[NUM_X_BINS][NUM_Y_BINS] = {{0}};
  char buffer[1024];
  int64_t i;
  if (max.params[j]==min.params[j]) return;
  if (max.params[k]==min.params[k]) return;
  double upbin_j = (double)NUM_X_BINS/(max.params[j]-min.params[j]);
  double upbin_k = (double)NUM_Y_BINS/(max.params[k]-min.params[k]);
  int64_t bin_x, bin_y;
  for (i=0; i<num_models; i++) {
    bin_x = (models[i].params[j]-min.params[j])*upbin_j;
    bin_y = (models[i].params[k]-min.params[k])*upbin_k;
    if (bin_x>=NUM_X_BINS) bin_x=NUM_X_BINS-1;
    if (bin_y>=NUM_Y_BINS) bin_y=NUM_Y_BINS-1;
    bins[bin_x][bin_y]++;
  }

  snprintf(buffer, 1024, "pdfs/dist_%"PRId64"_%"PRId64".dat", j, k);
  FILE *data = check_fopen(buffer, "w");
  for (bin_x=0; bin_x<NUM_X_BINS; bin_x++) {
    for (bin_y=0; bin_y<NUM_Y_BINS; bin_y++) {
      if (bins[bin_x][bin_y]>0)
	fprintf(data, "%lf %lf %lf\n", min.params[j] + (bin_x+0.5)*(max.params[j]-min.params[j])/(double)(NUM_X_BINS),
		min.params[k] + (bin_y+0.5)*(max.params[k]-min.params[k])/(double)(NUM_Y_BINS), bins[bin_x][bin_y]);
    }
  }
  fclose(data);
  fprintf(output, "%s %lf %lf %lf %lf\n", buffer, min.params[j], max.params[j], min.params[k], max.params[k]);
}
