#!/usr/bin/perl -w
package RedshiftUtil;
use Exporter 'import';
our @EXPORT_OK = qw(get_scales get_scales_at_redshifts find_redshift);
our %EXPORT_TAGS = ('all' => \@EXPORT_OK);

sub get_scales {
    my $dir = shift() || ".";
    my $fn = shift();
    my $reg = qr/a(\d\.\d+)/;
    if (defined $fn) {
	$reg = qr/^\Q$fn\E.*?a(\d\.\d+)/;
    }
    opendir DIR, $dir or die "Could not open directory $dir! ($!)\n";
    my %scales = map {$_ => 1} map { /$reg/ } readdir DIR;    
    closedir DIR;
    return sort { $a <=> $b} keys %scales;
}

sub find_redshift {
    my ($z, @scales) = @_;
    my $best_scale = $scales[0];
    my $best_z = 1/$best_scale - 1;
    for (@scales) {
	my $tz = 1/$_ - 1;
	if (abs($tz - $z) < abs($best_z - $z)) {
	    $best_scale = $_;
	    $best_z = $tz;
	}
    }
    return $best_scale;
}

sub get_scales_at_redshifts {
    my @scales = get_scales(@_);
    my @redshifts = @_[2..$#_];
    return map { find_redshift($_, @scales) } @redshifts;
}

1;
