#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "../check_syscalls.h"

double *autocorrelation(double *params, int64_t nwalkers, int64_t total_steps) {
  int64_t i, ss;
  double a1=0, a2=0, s1=0, s2=0, s12=0;
  int64_t max_step_size = total_steps/nwalkers;
  double *acor = NULL;
  check_realloc_s(acor, sizeof(double), max_step_size);
  acor[0] = 1;
  for (ss=1; ss<max_step_size; ss++) {
    int64_t max_i = total_steps - (ss*nwalkers);
    for (i=0; i<max_i; i++) {
      a1 += params[i];
      a2 += params[i+ss*nwalkers];
    }
    a1 /= (double)max_i;
    a2 /= (double)max_i;
    for (i=0; i<max_i; i++) {
      double d1 = params[i]-a1;
      s1 += d1*d1;
      double d2 = params[i+ss*nwalkers] - a2;
      s2 += d2*d2;
      s12 += d1*d2;
    }
    s1 /= (double)max_i;
    s2 /= (double)max_i;
    s12 /= (double)max_i;
    acor[ss] = (s1 > 0 && s2 > 0) ? s12/sqrt(s1*s2) : 1;
  }
  return acor;
}

void print_autocorrelation(double *params, int64_t nwalkers, int64_t total_steps, FILE *out) {
  int64_t i;
  int64_t max_step_size = total_steps/nwalkers;
  double *acor = autocorrelation(params, nwalkers, total_steps);
  for (i=1; i<max_step_size; i++) {
    fprintf(out, "%"PRId64" %f\n", i, acor[i]);
  }
  fprintf(out, "\n");
  free(acor);
}


int main(int argc, char **argv) {
  char buffer[1024];
  double *params = NULL;
  double p = 0;
  int64_t num_params = 0;
  while (fgets(buffer, 1024, stdin)) {
    if (sscanf(buffer, "%lf", &p)==1) {
      check_realloc_every(params, sizeof(double), num_params, 1000);
      params[num_params]=p;
      num_params++;
    }    
  }
  print_autocorrelation(params, 1, num_params, stdout);
}
