#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inttypes.h>
#include "../mt_rand.h"
#include "../check_syscalls.h"

#define DIM 46
#define TRIALS 1000

float *all_vecs = NULL;

float dot(float *v1, float *v2) {
  int64_t i;
  float s = 0;
  for (i=0; i<DIM; i++)
    s += v1[i]*v2[i];
  return s;
}

void rand_vec(float *v) {
  int64_t i;
  for (i=0; i<DIM; i++) v[i] = normal_random(0,1);
}

float est_grad(int64_t num_vec) {
  int64_t i, j, k;
  memset(all_vecs, 0, sizeof(float)*num_vec*DIM);
  float best_vec[DIM] = {0};
  for (i=0; i<num_vec; i++) {
    rand_vec(all_vecs+i*DIM);
    //Perpendicularize:
    for (j=0; j<i; j++) {
      float d = dot(all_vecs+i*DIM, all_vecs+j*DIM);
      for (k=0; k<DIM; k++) all_vecs[i*DIM+k] -= d*all_vecs[j*DIM+k];
    }
    //Make unit length
    float s = dot(all_vecs+i*DIM, all_vecs+i*DIM);
    s = 1.0/sqrt(s);
    for (k=0; k<DIM; k++) {
      all_vecs[i*DIM+k] *= s;
    }

    //Compute new best vector
    float bv_grad = best_vec[0];
    float new_grad = all_vecs[i*DIM];
    for (k=0; k<DIM; k++) {
      best_vec[k] = best_vec[k]*bv_grad + all_vecs[i*DIM+k]*new_grad;
    }

    //Make vector unit length
    s = dot(best_vec, best_vec);
    s = 1.0/sqrt(s);
    for (k=0; k<DIM; k++)
      best_vec[k] *= s;
  }
  return best_vec[0];
}

int main(void) {
  int64_t i, j;
  float t[DIM] = {0};
  check_realloc_s(all_vecs, sizeof(float)*DIM, DIM);
  
  r250_init(87L);
  for (i=1; i<DIM; i++) {
    float c=0;
    for (j=0; j<TRIALS; j++)
      c += est_grad(i);
    c /= (float)TRIALS;
    printf("%"PRId64" %f\n", i, c);
  }
}
