#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inttypes.h>
#include "../mt_rand.h"

#define NDIM 2

struct point {
  double pos[NDIM];
};

int64_t total_evals = 0;

double calc_chi2(double pos[NDIM]) {
  int64_t i;
  double chi2 = 0;
  for (i=0; i<NDIM; i++) {
    chi2 += pos[i]*pos[i];
  }
  total_evals++;
  return chi2;
}

double calc_chi2_step(double pos[NDIM], double v1[NDIM], double v2[NDIM], double x, double y) {
  double pos2[NDIM]={0};
  for (int64_t i=0; i<NDIM; i++) {
    pos2[i] = pos[i] + x*v1[i] + y*v2[i];
  }
  return calc_chi2(pos2);
}


int main(void) {
  int64_t i,j;
  struct point p = {{0}};
  p.pos[0] = 6;
  double chi2 = calc_chi2(p.pos);
  double step = 0.1;
  double move = 10.0;
  double v1[2] = {1,0};
  double v2[2] = {0,1};
  double orig_chi2 = chi2;
  double dir = -1;
  double last_atan = atan2(p.pos[1], p.pos[0]);
  int64_t total_loops=0;
  r250_init(87L);
  for (i=0; i<100000; i++) {
    double dcdx = (calc_chi2_step(p.pos, v1, v2, step*dir, 0)-chi2)/(dir*step);
    double dcdy = (calc_chi2_step(p.pos, v1, v2, 0, step*dir)-chi2)/(step*dir);

    //Move
    double abs_dc = sqrt(dcdx*dcdx+dcdy*dcdy);
    double a = -dcdy/abs_dc;
    double b = dcdx/abs_dc;

    double thismove = move*dr250()*2.0;
    p.pos[0] += a*thismove;
    p.pos[1] += b*thismove;

    chi2 = calc_chi2(p.pos);
    
    //Apply correction...
    j=0;
    while (fabs(chi2-orig_chi2) > 0.001 && j<10) {
      j++;
      dcdx = (calc_chi2_step(p.pos, v1, v2, step*dir*-1, 0)-chi2)/(step*dir*-1);
      dcdy = (calc_chi2_step(p.pos, v1, v2, 0, step*dir*-1)-chi2)/(step*dir*-1);
      double old_abs_dc = abs_dc;
      abs_dc = sqrt(dcdx*dcdx+dcdy*dcdy);

      double normalization = sqrt(abs_dc*old_abs_dc);
      //abs_dc = old_abs_dc; //
      double corr_step = (orig_chi2-chi2)/normalization; //*pow(new_chi2/orig_chi2, 0.25);
      p.pos[0] += corr_step*dcdx/abs_dc;
      p.pos[1] += corr_step*dcdy/abs_dc;
      chi2 = calc_chi2(p.pos);
      dir *= -1;
    }

    double new_atan = atan2(p.pos[1], p.pos[0]);
    if (new_atan < last_atan) { total_loops++; }
    last_atan=new_atan;
    if (!(j%2)) dir *= -1;
    printf("%f %f %f %f %f\n", p.pos[0], p.pos[1], dcdx, dcdy, chi2);
  }
  printf("#Total evals: %"PRId64"\n", total_evals);
  printf("#Total evals/loop: %lf\n", (double)total_evals/(double)total_loops);
  return 0;
}
