#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <math.h>
#include <inttypes.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <assert.h>
#include "../check_syscalls.h"
#define SAMPLE_CORR
#define SPLASHBACK
#include "../make_sf_catalog.h"
#include "../mt_rand.h"

#define FAST3TREE_TYPE struct catalog_galaxy
#include "../fast3tree.c"

#define REDSHIFT_ERROR 35.0
#define BOX_SIZE 250.0
#define NEIGHBOR_DSM 0.25
#define RANDOM_FACTOR 10
#define h 0.678

int64_t num_galaxies = 0;
struct catalog_galaxy *cg = NULL;

float min_dsm=0, max_dsm=0;
struct fast3tree *t=NULL;
struct fast3tree_results *res=NULL;

int64_t calc_ncounts(float *pos, float sm, float *r1, float *r2);
float proj_dist(float *pos1, float *pos2);

int main(int argc, char **argv) {
  int64_t i, j, k, new_length;
  if (argc<2) {
    printf("Usage: %s sfr_list.bin [min_mass max_mass proj_iso redshift_iso]\n", argv[0]);
    exit(1);
  }

  float scale = 1;
  char *scale_text = argv[1];
  while (scale_text[0] && (scale_text[0]<'0' || scale_text[0]>'9')) scale_text++;
  if (scale_text[0]) scale = atof(scale_text);
  printf("#Scale: %f\n", scale);
  printf("#Assumed cosmo: Om=0.307; h=%f\n", h);
  double H0 = 100.0*sqrt(0.307/pow(scale,3) + 0.693);
  printf("#H0: %f km/s/(Mpc/h)\n", H0);

  float min_mass = 10.5;
  float max_mass = 11;
  if (argc>2) min_mass = atof(argv[2]);
  if (argc>3) max_mass = atof(argv[3]);
  if (min_mass < 15) min_mass = pow(10, min_mass);
  if (max_mass < 15) max_mass = pow(10, max_mass);

  float proj_iso = 2;
  if (argc>4) proj_iso = atof(argv[4]);
  proj_iso *= h;
    
  float redshift_iso = 1000;
  if (argc>5) redshift_iso = atof(argv[5]);

  printf("#Assumed redshift error: %g km/s\n", REDSHIFT_ERROR);
  printf("#Assumed box size: %g Mpc/h\n", BOX_SIZE);
  printf("#Assumed neighbor DM: +/- %g dex\n", NEIGHBOR_DSM);
  printf("#Mass range: %g - %g\n", log10(min_mass), log10(max_mass));
  printf("#Projected isolation: %g Mpc/h\n", proj_iso);
  printf("#Redshift isolation: %g km/s\n", redshift_iso);
  printf("#PX PY RZ X Y Z VX VY VZ SM SFR Neighbors(0.3-0.4) HM UPID DeltaVmax_Rank SSFR\n");
  
  redshift_iso /= H0;
  float mass_threshold = min_mass *pow(10, -NEIGHBOR_DSM);
  r250_init(87L);
  
  int64_t num_cg = 0;
  struct catalog_galaxy *cg = NULL;
  
  struct catalog_halo *p = check_mmap_file(argv[1], 'r', &new_length);
  int64_t num_p = new_length / sizeof(struct catalog_halo);

  for (i=0; i<num_p; i++) {
    struct catalog_halo *tp = p+i;
    if (tp->flags & IGNORE_FLAG) continue;
    if (tp->obs_sm >= mass_threshold) {
      check_realloc_every(cg, sizeof(struct catalog_galaxy), num_cg, 1000);
      memcpy(cg[num_cg].pos, tp->pos, sizeof(float)*3);
      memcpy(cg[num_cg].opos, tp->pos, sizeof(float)*6);
      cg[num_cg].pos[2] += (normal_random(0,REDSHIFT_ERROR) + tp->pos[5])/H0;
      if (cg[num_cg].pos[2] < 0) cg[num_cg].pos[2] += BOX_SIZE;
      if (cg[num_cg].pos[2] > BOX_SIZE) cg[num_cg].pos[2] -= BOX_SIZE;
      cg[num_cg].lsm = tp->obs_sm;
      cg[num_cg].lsfr = tp->obs_sfr;
      cg[num_cg].hm = tp->mp;
      cg[num_cg].rank1 = tp->rank1;
      cg[num_cg].upid = tp->upid;

      cg[num_cg].ssfr = 0;
      if (tp->sm>0) cg[num_cg].ssfr = tp->obs_sfr/tp->obs_sm;
      
      num_cg++;
    }
  }

  t = fast3tree_init(num_cg, cg);
  _fast3tree_set_minmax(t, 0, BOX_SIZE);
  res = fast3tree_results_init();

  float r1[3] = {proj_iso,proj_iso,redshift_iso};
  float r2[3] = {4*h,4*h,1000.0/H0};

  min_dsm = pow(10, -NEIGHBOR_DSM);
  max_dsm = pow(10, NEIGHBOR_DSM);
  FILE *rf = check_fopen("splashback_randoms.txt", "w");
  FILE *rf_noenv = check_fopen("splashback_randoms_noenv.txt", "w");

  for (i=0; i<num_cg; i++) {
    if (cg[i].lsm <= min_mass) continue;
    if (cg[i].lsm >= max_mass) continue;
    int64_t ncounts = calc_ncounts(cg[i].pos, cg[i].lsm, r1, r2);
    if (ncounts > 9 || ncounts < 0) continue;
    for (k=0; k<3; k++) printf("%f ", cg[i].pos[k]);
    for (k=0; k<6; k++) printf("%f ", cg[i].opos[k]);
    printf("%g %g %"PRId64" %e %"PRId64" %f %e\n", cg[i].lsm, cg[i].lsfr, ncounts, cg[i].hm, cg[i].upid, cg[i].rank1, cg[i].ssfr);

    int64_t num_randoms = 0, num_noenv = 0;
    while (num_randoms < RANDOM_FACTOR) {
      float pos[3];
      for (j=0; j<3; j++) pos[j] = dr250()*BOX_SIZE;
      ncounts = calc_ncounts(pos, cg[i].lsm, r1, r2);
      if (num_noenv < RANDOM_FACTOR) {
	for (k=0; k<3; k++) fprintf(rf_noenv, "%f ", pos[k]);
	for (k=0; k<3; k++) fprintf(rf_noenv, "%f ", pos[k]);
	for (k=0; k<3; k++) fprintf(rf_noenv, "0 ");
	fprintf(rf_noenv, "%g %g %"PRId64"\n", cg[i].lsm, cg[i].lsfr, ncounts);
	num_noenv++;
      }
      if (ncounts > 9 || ncounts < 0) continue;
      for (k=0; k<3; k++) fprintf(rf, "%f ", pos[k]);
      for (k=0; k<3; k++) fprintf(rf, "%f ", pos[k]);
      for (k=0; k<3; k++) fprintf(rf, "0 ");
      fprintf(rf, "%g %g %"PRId64"\n", cg[i].lsm, cg[i].lsfr, ncounts);
      num_randoms++;
    }
  }
  munmap(p, new_length);
  fclose(rf);
  fclose(rf_noenv);
  return 0;
}

float proj_dist(float *pos1, float *pos2) {
  int64_t k;
  float ds=0, dx;
  for (k=0; k<2; k++) {
    dx = fabs(pos1[k]-pos2[k]);
    if (dx>BOX_SIZE/2.0) dx = BOX_SIZE-dx;
    ds += dx*dx;
  }
  return sqrtf(ds);
}

int64_t calc_ncounts(float *pos, float sm, float *r1, float *r2) {
  int pdims[3] = {1,1,1};
  int64_t j;
  fast3tree_find_ellipsebox_periodic(t, res, pos, r1, pdims);
  for (j=0; j<res->num_points; j++)
    if (res->points[j]->lsm > sm) return -1;
  fast3tree_find_ellipsebox_periodic(t, res, pos, r2, pdims);
  int64_t ncounts = 0;
  for (j=0; j<res->num_points; j++) {
    if (res->points[j]->lsm < min_dsm*sm) continue;
    if (res->points[j]->lsm > max_dsm*sm) continue;
    if (proj_dist(res->points[j]->pos, pos) < 0.3) continue;
    ncounts++;
  }
  return ncounts;
}
