#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <math.h>
#include <inttypes.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <assert.h>
#include "../check_syscalls.h"
#include "../make_sf_catalog.h"
#include "../mt_rand.h"
#include "../stringparse.h"

#define FAST3TREE_TYPE struct catalog_galaxy
#define FAST3TREE_DIM 3
#include "../fast3tree.c"

#define BOX_SIZE 250.0
#define h 0.678

//In Mpc
#define RAD_MIN -2
#define RAD_MAX 1
#define RAD_BPDEX 5
#define RAD_BINS ((RAD_MAX-RAD_MIN)*RAD_BPDEX+1)

int64_t num_all=0, num_sf=0, num_q=0;
double all_bins[RAD_BINS]={0}, sf_bins[RAD_BINS]={0}, q_bins[RAD_BINS]={0}, ssf_bins[RAD_BINS]={0}, sq_bins[RAD_BINS]={0};

int64_t num_galaxies = 0;
struct catalog_galaxy *cg = NULL;

struct fast3tree *t=NULL;
struct fast3tree_results *res=NULL;

float proj_dist(float *pos1, float *pos2);
void renorm(double *bins, int64_t num) {
  int64_t i;
  if (!(num>0)) return;
  for (i=0; i<RAD_BINS; i++) {
    float rmin = pow(10, RAD_MIN+(double)i/(double)RAD_BPDEX);
    float rmax = pow(10, RAD_MIN+(double)(i+1)/(double)RAD_BPDEX);
    float area = (4.0*M_PI/3.0)*(rmax*rmax*rmax-rmin*rmin*rmin);
    //float rmed = 0.5*sqrt(rmax*rmax + rmin*rmin);
    bins[i] /= ((double)num)*area;
  }
}


int main(int argc, char **argv) {
  int64_t i, new_length;
  if (argc<3) {
    printf("Usage: %s sfr_list.bin splashback.txt [min_sm]\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  float scale = 1;
  char *scale_text = argv[1];
  while (scale_text[0] && (scale_text[0]<'0' || scale_text[0]>'9')) scale_text++;
  if (scale_text[0]) scale = atof(scale_text);
  printf("#Scale: %f\n", scale);
  printf("#Assumed cosmo: h=0.678\n");

  float min_sm = 9;
  if (argc>3) min_sm = atof(argv[3]);
  if (min_sm < 15) min_sm = pow(10, min_sm);

  printf("#Assumed box size: %g Mpc/h\n", BOX_SIZE);


  printf("#R All_dens SF_Host_dens Q_Host_dens Sat_SF_dens Sat_Q_dens\n");
  
  int64_t num_cg = 0;
  struct catalog_galaxy *cg = NULL;
  
  struct catalog_halo *p = check_mmap_file(argv[1], 'r', &new_length);
  int64_t num_p = new_length / sizeof(struct catalog_halo);

  for (i=0; i<num_p; i++) {
    struct catalog_halo *tp = p+i;
    if (tp->flags & IGNORE_FLAG) continue;
    if (tp->obs_sm >= min_sm) {
      check_realloc_every(cg, sizeof(struct catalog_galaxy), num_cg, 1000);
      memcpy(cg[num_cg].pos, tp->pos, sizeof(float)*3);
      cg[num_cg].lsm = tp->obs_sm;
      cg[num_cg].lsfr = tp->obs_sfr;
      num_cg++;
    }
  }
  munmap(p, new_length);


  t = fast3tree_init(num_cg, cg);
  _fast3tree_set_minmax(t, 0, BOX_SIZE);
  res = fast3tree_results_init();

  FILE *in = check_fopen(argv[2], "r");
  char buffer[2048];
  SHORT_PARSETYPE;
  float sm, sfr;
  float pos[3];
  enum parsetype dtypes[11] = {K, K, K, F, F, F, K, K, K, F, F};
  void *data[11] = {NULL, NULL, NULL, pos, pos+1, pos+2, NULL, NULL, NULL, &sm, &sfr};

  float max_r = pow(10, RAD_MAX+1.0/(double)RAD_BPDEX)*h;
  float min_r = pow(10, RAD_MIN)*h;
  while (fgets(buffer, 2048, in)) {
    if (buffer[0]=='#') continue;
    if (stringparse(buffer, data, dtypes, 11)!=11) continue;
    fast3tree_find_sphere_periodic(t, res, pos, max_r);

    num_all++;
    if (sfr > sm*1e-11) num_sf++;
    else num_q++;
    
    for (i=0; i<res->num_points; i++) {
      float dist = proj_dist(pos, res->points[i]->pos)/h;
      if (dist < min_r/h) continue;
      dist = log10(dist);
      int64_t dist_bin = (dist-RAD_MIN)*RAD_BPDEX;
      all_bins[dist_bin]++;
      if (sfr > sm*1e-11) sf_bins[dist_bin]++;
      else q_bins[dist_bin]++;
      if (res->points[i]->lsfr > res->points[i]->lsm*1e-11) ssf_bins[dist_bin]++;
      else sq_bins[dist_bin]++;
    }
  }

  renorm(all_bins, num_all);
  renorm(sf_bins, num_sf);
  renorm(q_bins, num_q);
  renorm(ssf_bins, num_all);
  renorm(sq_bins, num_all);

  for (i=0; i<RAD_BINS; i++) {
    float rmin = pow(10, RAD_MIN+(double)i/(double)RAD_BPDEX);
    float rmax = pow(10, RAD_MIN+(double)(i+1)/(double)RAD_BPDEX);
    float rmed = 0.5*cbrt(rmax*rmax*rmax + rmin*rmin*rmin);
    printf("%f %g %g %g %g %g\n", rmed, all_bins[i], sf_bins[i], q_bins[i], ssf_bins[i], sq_bins[i]);
  }
  return 0;
}

float proj_dist(float *pos1, float *pos2) {
  int64_t k;
  float ds=0, dx;
  for (k=0; k<3; k++) {
    dx = fabs(pos1[k]-pos2[k]);
    if (dx>BOX_SIZE/2.0) dx = BOX_SIZE-dx;
    ds += dx*dx;
  }
  return sqrtf(ds);
}

